/* YOODA : Yet another Object Oriented Database
   Copyright (C) 1994  Eric Abecassis

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */


/*****************************************************************************
*
*  MODULE :             TableOfBits
*  SCOPE :              Allocation through table of bits
*  FILE :               tableofbits.cc
*  AUTHORS :            E. Abecassis
*  CREATION DATE :      August 1, 1993
*  COPYRIGHTS :        (C) E. Abecassis, 1993
*  HISTORY :
*
******************************************************************************/

/*=
[DESCRIPTION :]

   This class is a tool similar to fd_set. It is a table of bits. Each bit
   can be set or clear. Interface of the class allows to set, clear, or test
   sequence of bits. A method allows to find the first sequence of clear bits.
   This method can be used to do allocation.

=*/

/* IMPORTS *******************************************************************/

#ifdef __GNUG__
#pragma implementation
#endif

#ifndef _tableofbits_HXX
#include "tableofbits.hxx"
#endif

#ifndef _memory_H
#define _memory_H
#include <memory.h>
#endif


/*= EXTERNAL FUNCTION =======================================================*/
int TableOfBits::set(int pos, int len)
/*=
[DESCRIPTION :]
  This method sets 'len' bits in the table of bits. All the bits must be
  contains in the limit of the table. If not, the method returns the bits which
  have been set up to the limit
  (IN)
    int pos : position of the first bit we want to set
    int len : length of the sequence of bits we want to set
  (OUT)
    return int : number of set bits 
=*/
{
  register int cpt = 0 ;
  for (; (len > 0) && (pos < this->_size); len--, pos++, cpt++) {
    this->set(pos) ;
  }
  return cpt ;
}


/*= EXTERNAL FUNCTION =======================================================*/
int TableOfBits::clear(int pos, int len)
/*=
[DESCRIPTION :]
  This method clears 'len' bits in the table of bits. All the bits must be
  contains in the limit of the table. If not, the method returns the bits which
  have been set up to the limit
  (IN)
    int pos : position of the first bit we want to clear
    int len : length of the sequence of bits we want to clear
  (OUT)
    return int : number of cleared bits 
=*/
{
  register int cpt = 0 ;
  for (; (len > 0) && (pos < this->_size); len--, pos++, cpt++) {
    this->clear(pos) ;
  }
  return cpt ;
}


/*= EXTERNAL FUNCTION =======================================================*/
void TableOfBits::zero()
/*=
[DESCRIPTION :]
  Fill the table with clear bits
=*/
{
  memset(_tab, 0, _size / NUMBITS(char)) ;
}


/*= EXTERNAL FUNCTION =======================================================*/
void TableOfBits::one()
/*=
[DESCRIPTION :]
  Fill the table with clear bits
=*/
{
  memset(_tab, 0xff, _size / NUMBITS(char)) ;
}


/*= EXTERNAL FUNCTION =======================================================*/
int TableOfBits::isset(int pos, int len) const
/*=
[DESCRIPTION :]
  Test if the sequence defined by 'pos' and 'len' is set. Return the number
  of set bits found in the sequence.
  (IN)
    int pos : position of the first bit of the sequence
    int len : length of the sequence
  (OUT)
    return int : number of set bits found 
=*/
{
  register int cpt = 0 ;
  for (; (len > 0) && (pos < this->_size); len--, pos++, cpt++) {
    if (!this->isset(pos)) {
      return cpt ;
    }
  }
  return cpt ;
}


/*= EXTERNAL FUNCTION =======================================================*/
int TableOfBits::isclear(int pos, int len) const
/*=
[DESCRIPTION :]
  Test if the sequence defined by 'pos' and 'len' is clear. Return the number
  of cleared bits found in the sequence.
  (IN)
    int pos : position of the first bit of the sequence
    int len : length of the sequence
  (OUT)
    return int : number of clear bits found 
=*/
{
  register int cpt = 0 ;
  for (; (len > 0) && (pos < this->_size); len--, pos++, cpt++) {
    if (this->isset(pos)) {
      return cpt ;
    }
  }
  return cpt ;
}


/*= EXTERNAL FUNCTION =======================================================*/
int TableOfBits::look(int pos, int len) const 
/*=
[DESCRIPTION :]
  This method is used to search a sequence of 'len' clear bits in the table.
  To optimize the search, bits are scanned by 32 elements using integer.
  When we find the first integer which contains a clear bit, we try to find
  if it corresponds to a clear sequence of searched size. Note that the
  position where search must began is given as parameter. It returns -1 if
  no sequence are found or the position of the first bit of the search
  sequence.
  (IN)
    int pos : position where the search begins
    int len : length of the search sequence
  (OUT)
    return int : position of the search sequence or -1 if not
=*/
{
  register int sizeint = _size / NUMBITS(int) ;
  register int* tabint = (int*)_tab ;
  register int curr = nextclear(pos) ;
  while (curr < _size) {
    if ( (this->isset(curr)) && ((curr % NUMBITS(int)) == 0)) {
      for (register int i = (curr / NUMBITS(int));
	   (i < sizeint) && (tabint[i] == 0xffffffff) ; i++) ;
      curr = nextclear(i * NUMBITS(int)) ;
      continue ;
    }
    
    register int numfree = this->isclear(curr, len) ;
    if (numfree == len) {
      return curr ;
    } else {
      curr = nextclear(curr+numfree) ;
    }
  }
  return -1 ;
} 


/*= INTERNAL FUNCTION =======================================================*/
int TableOfBits::nextclear(int pos) const
/*=
[DESCRIPTION :]
  Return the next clear position from position 'pos' up to the next limit
  of 32 bits.
  (IN)
    int pos : first search position
  (OUT)
    return int : return the first clear bit or the next limit of 32 bits.
=*/
{
  do {
    if (this->isclear(pos)) {
      return pos ;
    } else {
      pos++ ;
    }
  } while ( (pos < _size) && ((pos % NUMBITS(int)) != 0) ) ;
  return pos ;
}


/*= EXTERNAL FUNCTION =======================================================*/
TableOfBits& TableOfBits::operator=(const TableOfBits& t)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  this->_doalloc = 1 ;
  this->_size = t._size ;
  this->_tab = new char[_size/NUMBITS(char)] ;
  memcpy(this->_tab, t._tab, _size/NUMBITS(char)) ;
  return *this ;
}






